from datetime import datetime
from typing import Optional
import pytz

import dateparser

def parse(value: Optional[str]) -> Optional[datetime]:
    if not value:
        return None
    x = dateparser.parse(value)
    return x.replace(tzinfo=pytz.UTC)


def parse_to_json(value: Optional[datetime]) -> Optional[str]:
    if not value or not isinstance(value, datetime):
        return None
    return value.isoformat()



from rich.console import Console

# ====================================================================================
#
# If you want to use the console object to print stuff in your script
# either use the corresponding log level method (verbose, info, critical)
# or use .log() as it will always output regardless of the configured loglevel.
#
# ====================================================================================


class MASZLogLevel:
    VERBOSE = 0
    INFO = 1
    CRITICAL = 2

class WithDummy:
    def __enter__(self, *args):
        pass
    def __exit__(self, *args):
        pass

class MASZConsole(Console):
    def __init__(self, log_level: MASZLogLevel = MASZLogLevel.INFO):            
        super().__init__()
        self.log_level = log_level
        
    def verbose(self, string: str):
        if self.log_level == 0:
            self.log(f"[bright_black][V][/bright_black] {string}")

    def info(self, string: str):
        if self.log_level <= 1:
            self.log(f"[bright_black][[white]I[/white]][/bright_black] {string}")

    def critical(self, string: str):
        self.log(f"[bright_black][[bright_red]C[/bright_red]][/bright_black] {string}")

    def info_status(self, string: str):
        if self.log_level <= 1:
            return self.status(string)
        return WithDummy()

    def verbose_status(self, string: str):
        if self.log_level == 0:
            return self.status(string)
        return WithDummy()

console = MASZConsole()



class GuildConfig:
    def __init__(self, **kwargs) -> None:
        self.id = kwargs.get("id")
        self.guild_id = kwargs.get("guildId")
        self.mod_roles = kwargs.get("modRoles", [])
        self.admin_roles = kwargs.get("adminRoles", [])
        self.muted_roles = kwargs.get("mutedRoles", [])
        self.dm_notification = kwargs.get("modNotificationDM", False)
        self.public_webhook = kwargs.get("modPublicNotificationWebhook", '')
        self.internal_webhook = kwargs.get("modInternalNotificationWebhook", '')
        self.strict_permission_check = kwargs.get("strictModPermissionCheck", False)
        self.execute_whois_on_join = kwargs.get("executeWhoisOnJoin", False)

    def __str__(self) -> str:
        return self.guild_id



from masz.helpers import parse_dt_from_json

from .discord_user import DiscordUser

class DiscordMember:
    def __init__(self, **kwargs) -> None:
        self.user = DiscordUser(**kwargs.get("user"))
        self.nick = kwargs.get("nick")
        self.roles = kwargs.get("roles", [])
        self.joined_at = parse_dt_from_json(kwargs.get("type"))

    def __str__(self) -> str:
        return self.nick if self.nick else f"{self.user.username}#{self.user.discriminator}"



from requests import Response

from .console import console, MASZLogLevel
from .request_adapter import MASZRequestAdapter
from .exceptions import MASZBaseException, MASZLoginFailure
from .obj import DiscordUser


SUPPORTED_API_VERSION = 1

class MASZClient(MASZRequestAdapter):
    def __init__(self, url: str, token: str, api_version: int = 1, log_level: MASZLogLevel = MASZLogLevel.INFO) -> None:
        super().__init__(url, token, api_version)
        self._url = url
        self._token = token
        self._api_version = api_version
        console.log_level = log_level

        try:
            with console.info_status(f"[bold green]Connecting to MASZ at {url} ...") as status:
                self.server_health = self.get_current_health()
                if self.server_health.status.lower() == 'ok':
                    if self.server_health.response_time < 200:
                        console.info(f":white_check_mark: Ping [bright_green]'{self.server_health.status.upper()}' {self.server_health.response_time}[/bright_green] ms.")
                    else:
                        console.info(f":white_check_mark: Ping [bright_green]'{self.server_health.status.upper()}'[/bright_green][bright_yellow] {self.server_health.response_time}[/bright_yellow] ms.")
                else:
                    console.info(f":white_check_mark: Ping [bright_red]'{self.server_health.status.upper()}' {self.server_health.response_time}[/bright_red] ms.")

        except MASZBaseException as e:
            console.critical(f":exclamation: [red]MASZ API at {url} seems to be unhealthy or unreachable.")
            exit(1)

        try:
            with console.info_status(f"[bold green]Checking API version...") as status:
                self.server_version = self.get_version()
                if self.server_version.wrapper_version is None:
                    console.critical(f":exclamation: This API [red]may not be compatible[/red]. You need to use MASZ v1.12.1 or newer!")
                else:
                    if SUPPORTED_API_VERSION >= self.server_version.wrapper_version:
                        console.info(f":white_check_mark: API version {self.server_version.wrapper_version} is [bright_green]compatible[/bright_green].")
                    else:
                        console.critical(f":exclamation: API version {self.server_version.wrapper_version} [red]may not be compatible[/red]. This package only supports versions until {SUPPORTED_API_VERSION}. Try to upgrade this python package.")

        except MASZBaseException as e:
            console.critical(f":exclamation: [red]Failed to fetch API version.")

        try:
            with console.info_status(f"[bold bright_green]Logging in...") as status:
                self.me = self.get_current_user()
                console.info(f":white_check_mark: [bright_green]Logged in[/bright_green] to MASZ at {url} as [bright_blue]{self.me}[/bright_blue].")

        except MASZBaseException as e:
            console.critical(f":exclamation: [red]Failed to login.")
            exit(1)

    def __str__(self) -> str:
        return f"MASZ at {self._url} | {self.server_health}\n{self.server_version}"



from .discord_user import DiscordUser
from .app_user import AppUser
from .status import Status
from .version import Version
from .adminstats import Adminstats
from .guild import DiscordGuild
from .role import DiscordRole
from .guildconfig import GuildConfig
from .channel import DiscordChannel
from .member import DiscordMember
from .modcase import Modcase
from .punishment_type import PunishmentType
from .creation_type import CreationType
from .comment import Comment
from .motd import Motd
from .usernote import UserNote
from .usermap import UserMap



from typing import List
from datetime import datetime

from masz.helpers import parse_dt_from_json, parse_dt_to_json, parse_enum_to_json
from .punishment_type import PunishmentType
from .creation_type import CreationType
from .comment import Comment

class Modcase:
    comments: List[Comment]

    def __init__(self, **kwargs) -> None:
        self.id = kwargs.get("id")
        self.case_id = kwargs.get("caseId")
        self.guild_id = kwargs.get("guildId")
        self.title = kwargs.get("title")
        self.description = kwargs.get("description")
        self.user_id = kwargs.get("userId")
        self.username = kwargs.get("username")
        self.discriminator = kwargs.get("discriminator")
        self.nickname = kwargs.get("nickname")
        self.mod_id = kwargs.get("modId")
        self.created_at = parse_dt_from_json(kwargs.get("createdAt"))
        self.occured_at = parse_dt_from_json(kwargs.get("occuredAt"))
        self.last_edited_at = parse_dt_from_json(kwargs.get("lastEditedAt"))
        self.last_edited_by_mod_id = kwargs.get("lastEditedByModId")
        self.punishment = kwargs.get("punishment")
        self.labels = kwargs.get("labels", [])
        self.others = kwargs.get("others")
        self.valid = kwargs.get("valid", False)
        self.creation_type = kwargs.get("creationType", CreationType.DEFAULT)
        self.punishment_type = kwargs.get("punishmentType", PunishmentType.NONE)
        self.punished_until = parse_dt_from_json(kwargs.get("punishedUntil"))
        self.punishment_active = kwargs.get("punishmentActive", False)
        self.allow_comments = kwargs.get("allowComments", False)
        self.locked_by_user_id = kwargs.get("lockedByUserId")
        self.locked_at = parse_dt_from_json(kwargs.get("lockedAt"))
        self.marked_to_delete_at = parse_dt_from_json(kwargs.get("markedToDeleteAt"))
        self.deleted_by_user_id = kwargs.get("deletedByUserId")
        if kwargs.get("comments", []):
            self.comments = [Comment(**x) for x in kwargs.get("comments", [])]
        else:
            self.comments = []

    def __str__(self) -> str:
        return f"#{self.case_id} {self.title}"

    def to_dict(self) -> dict:
        return {
            "id": self.id,
            "caseId": self.case_id,
            "guildId": self.guild_id,
            "title": self.title,
            "description": self.description,
            "userId": self.user_id,
            "username": self.username,
            "discriminator": self.discriminator,
            "nickname": self.nickname,
            "modId": self.mod_id,
            "createdAt": parse_dt_to_json(self.created_at),
            "occuredAt": parse_dt_to_json(self.occured_at),
            "lastEditedAt": parse_dt_to_json(self.last_edited_at),
            "lastEditedByModId": self.last_edited_by_mod_id,
            "punishment": self.punishment,
            "labels": self.labels,
            "others": self.others,
            "valid": self.valid,
            "creationType": parse_enum_to_json(self.creation_type),
            "punishmentType": parse_enum_to_json(self.punishment_type),
            "punishedUntil": parse_dt_to_json(self.punished_until),
            "punishmentActive": self.punishment_active,
            "allowComments": self.allow_comments,
            "lockedByUserId": self.locked_by_user_id,
            "lockedAt": parse_dt_to_json(self.locked_at),
            "markedToDeleteAt": parse_dt_to_json(self.marked_to_delete_at),
            "deletedByUserId": self.deleted_by_user_id,
            "comments": [x.to_dict() for x in self.comments]
        }



import time
from typing import Union

from .masz_guild import MASZGuildAPI
from .masz_request import MASZRequestHandler
from .discord_api import MASZDiscordAPI
from .exceptions import *
from .obj import *


class MASZRequestAdapter:
    def __init__(self, url: str, token: str, api_version: int = 1, header: str = "Authorization", header_prefix: str = "Bearer ") -> None:
        self.request_handler = MASZRequestHandler(url, token, api_version, header, header_prefix)
        self.discord = MASZDiscordAPI(self.request_handler)

    def get_current_user(self) -> AppUser:
        return self.discord.get_current_user()

    def get_current_health(self) -> Status:
        start = time.perf_counter()
        r = self.request_handler.request("GET", "/health", headers={'Accept': 'application/json'})
        request_time = time.perf_counter() - start
        return Status(round(request_time*1000, 2) , **r.json())

    def get_version(self) -> Version:
        r = self.request_handler.request("GETSTATIC", "/static/version.json")
        return Version(**r.json())

    def get_adminstats(self) -> Adminstats:
        r = self.request_handler.request("GET", "/meta/adminstats")
        return Adminstats(**r.json())

    def guild(self, guild_id: Union[str, int]) -> MASZGuildAPI:
        return self.get_guild(guild_id=guild_id)

    def get_guild(self, guild_id: Union[str, int]) -> MASZGuildAPI:
        return MASZGuildAPI(self.request_handler, guild_id)



class DiscordChannel:
    def __init__(self, **kwargs) -> None:
        self.id = kwargs.get("id")
        self.name = kwargs.get("name")
        self.type = kwargs.get("type")

    def __str__(self) -> str:
        return self.name



from masz.obj.modcase import Modcase
from typing import Union

from .console import console
from .masz_request import MASZRequestHandler
from .exceptions import *
from .obj import *


class MASZUserMapAPI(UserMap):
    def __init__(self, request_handler: MASZRequestHandler, data: dict) -> None:
        self.request_handler = request_handler
        super().__init__(**data)

    def delete(self) -> bool:
        try:
            r = self.request_handler.request("DELETE", f"/guilds/{self.guild_id}/usermap/{self.id}")
        except MASZBaseException as e:
            console.verbose(f"Failed to delete usermap {e}")
            return False
        return r.status_code == 200

    def update(self, **fields) -> bool:
        for k, v in fields.items():
            setattr(self, k, v)
        try:
            r = self.request_handler.request("PUT", f"/guilds/{self.guild_id}/usermap",
                                                json_body=self.to_dict())
        except MASZBaseException as e:
            console.verbose(f"Failed to update usermap {e}")
            return False
        if r.status_code == 200:
            super().__init__(**r.json())
        return r.status_code == 200



from masz.obj.modcase import Modcase
from typing import Union

from masz.helpers import parse_dt_to_json
from .console import console
from .masz_request import MASZRequestHandler
from .exceptions import *
from .obj import *


class MASZModcaseAPI(Modcase):
    def __init__(self, request_handler: MASZRequestHandler, data: dict) -> None:
        self.request_handler = request_handler
        super().__init__(**data)

    def delete(self, send_notification: bool = True, force_delete: bool = False) -> bool:
        try:
            r = self.request_handler.request("DELETE", f"/modcases/{self.guild_id}/{self.case_id}", {'sendNotification': send_notification, 'forceDelete': force_delete})
        except MASZBaseException as e:
            console.verbose(f"Failed to delete modcase {e}")
            return False
        return r.status_code == 200

    def update(self, handle_punishment: bool = True, send_notification: bool = True, announce_dm: bool = True, **fields) -> bool:
        for k, v in fields.items():
            setattr(self, k, v)
        try:
            r = self.request_handler.request("PUT", f"/modcases/{self.guild_id}/{self.case_id}",
                                                params={'sendNotification': send_notification, 'handlePunishment': handle_punishment, 'announceDm': announce_dm},
                                                json_body=self.to_dict())
        except MASZBaseException as e:
            console.verbose(f"Failed to update modcase {e}")
            return False
        if r.status_code == 200:
            super().__init__(**r.json())
        return r.status_code == 200

    def post_comment(self, msg: Union[Comment, str]) -> bool:
        data = {
            "message": msg
        }
        if isinstance(msg, Comment):
            data = msg.to_dict()
        try:
            r = self.request_handler.request("POST", f"/modcases/{self.guild_id}/{self.case_id}/comments", json_body=data)
        except MASZBaseException as e:
            console.verbose(f"Failed to post comment {e}")
            return False

        if r.status_code == 201:  # refresh comments
            refresh = self.request_handler.request("GET", f"/modcases/{self.guild_id}/{self.case_id}")
            super().__init__(**refresh.json())

        return r.status_code == 201

    def delete_comment(self, comment: Union[Comment, str, int]) -> bool:
        comment_id = comment
        if isinstance(comment, Comment):
            comment_id = comment.id
        try:
            r = self.request_handler.request("DELETE", f"/modcases/{self.guild_id}/{self.case_id}/comments/{comment_id}")
        except MASZBaseException as e:
            console.verbose(f"Failed to delete comment {e}")
            return False
        if r.status_code == 200:
            self.comments = [x for x in self.comments if str(x.id)!=str(comment_id)]
        return r.status_code == 200

    def lock_comments(self) -> bool:
        r = self.request_handler.request("POST", f"/modcases/{self.guild_id}/{self.case_id}/lock")
        return r.status_code == 200

    def unlock_comments(self) -> bool:
        r = self.request_handler.request("DELETE", f"/modcases/{self.guild_id}/{self.case_id}/lock")
        return r.status_code == 200
    
    def restore(self) -> bool:
        r = self.request_handler.request("DELETE", f"/guilds/{self.guild_id}/bin/{self.case_id}/restore")
        return r.status_code == 200

    def delete_from_bin(self) -> bool:
        r = self.request_handler.request("DELETE", f"/guilds/{self.guild_id}/bin/{self.case_id}/delete")
        return r.status_code == 200



from .discord_user import DiscordUser

class AppUser:
    def __init__(self, **kwargs) -> None:
        self.discord_user = DiscordUser(**kwargs.get("discordUser"))
        self.is_siteadmin = bool(kwargs.get("isAdmin"))

    def __str__(self) -> str:
        if self.is_siteadmin:
            return f"{self.discord_user.username}#{self.discord_user.discriminator} (siteadmin)"
        return f"{self.discord_user.username}#{self.discord_user.discriminator}"
